/* fraglets0.15/src/cross.c

  parses genotypes (one per line, stdin), pairs them randomly and
  outputs the resulting two crossover genotypes (stdout)

  usage: cross [-d debugflag] [lim]
    lim  number of crossovers to compute (default: 1)

  Copyright (C) 2003-2004 <christian.tschudin@unibas.ch>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING

  History:

  20040703 doing crossover at the same position, instead of picking two
           random places in the two parents
  20030803 created, but it's a shame that we do not free memory
  
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/timeb.h>

#define MIN(a,b) ((a)<(b) ? (a) : (b))

struct frag {
	int nof;      // no of symbol fields
	char *cp;     // fraglet start (inside the genotype str)
	char **sym;
};

int gtcnt;
struct frag ***genotypes;


static int fragcmp(const void *a, const void *b)
{
	struct frag *f1 = *(struct frag**)a, *f2 = *(struct frag**)b;

	return strcmp(f1->cp, f2->cp);
}


void parse_genotype(char *cp)
{
	struct frag **fraglist, *f;
	int i, cnt = 0;

	fraglist = calloc(1, sizeof(struct frag*));
	if (!fraglist)
		return;

	// decompose genotype into fraglet strings
	while (*cp && *cp != '/')
		cp++;
	cp = strtok(cp, "/");
	while (cp) {
		char *cp2 = strchr(cp, ']');
		if (cp2) {
			f = calloc(1, sizeof(*f));
			*cp2 = '\0';
			f->cp = malloc(strlen(cp) + 1);;
			strcpy(f->cp, cp);

			// expand fraglet count by adding fraglets:
			for (i = atoi(cp2+1); i > 0; i--) {
				fraglist[cnt++] = f;
				fraglist = realloc(fraglist,
						   (cnt+1) * sizeof(*f));
				fraglist[cnt] = 0;
			}
		}
		cp = strtok(NULL, "/");
	}

	// sort genotype's fraglet strings:
//	qsort(fraglist, cnt, sizeof(struct frag *), fragcmp);

	// decompose fraglet strings into symbols:
	for (i = 0; i < cnt; i++) {
		f = fraglist[i];
		if (!f->sym) { // only parse symbol sequences once
			for (cp = strtok(f->cp, "[:]"); cp;
					     cp = strtok(NULL, "[:]")) {
				f->nof++;
				f->sym = realloc(f->sym,
						 f->nof * sizeof(char*));
				f->sym[f->nof - 1] = cp;
//				printf(" %s", cp);
			}
/*
		} else {
			for (j = 0; j < f->nof; j++) {
				printf(" %s", f->sym[j]);
			}
*/
		}
//		printf("\n");
	}

	// append new genotype
	genotypes = realloc(genotypes, ++gtcnt * sizeof(struct frag **));
	genotypes[gtcnt - 1] = fraglist;
}


void output(struct frag **fpp)
{
	int i, j, cnt;
	struct frag *f;

	// build string representation for each fraglet
	for (cnt = 0; fpp[cnt]; cnt++)
		fpp[cnt]->cp = 0;
	for (cnt = 0; fpp[cnt]; cnt++) {
		if (fpp[cnt]->cp)
			continue;
		f = fpp[cnt];
		for (i = 0, j = 2; i < f->nof; i++)
			j += 1 + strlen(f->sym[i]);
		f->cp = malloc(j);
		j = sprintf(f->cp, "%s", f->sym[0]); // node name
		for (i = 1; i < f->nof; i++)
			j += sprintf(f->cp + j,
				     i==1 ? "[%s" : ":%s", f->sym[i]);
	}

	// sort
//	qsort(fpp, cnt, sizeof(struct frag *), fragcmp);

	// print complete genotype
	printf("g ");
	for (i = 0, f = 0; i < cnt; i++) {
		f = fpp[i];
		printf("/%s", f->cp);
		// determine duplicates:
//		for (j = 1; fpp[i+1] && !strcmp(f->cp, fpp[i+1]->cp); i++)
		for (j = 1; fpp[i+1] &&
			(f == fpp[i+1] || !strcmp(f->cp, fpp[i+1]->cp)); i++)
			j++;
		printf("]%d", j);
	}
	printf("/\n");
}


void docross(struct frag **f1, struct frag **f2)
{
	int n1, n2, c1, c2, ca, cb;
	struct frag **fa, **fb, **fx, **fy, *f;

	// count total number of symbols in each genotype
	for (fa = f1, n1 = 0; *fa; fa++)
		n1 += (*fa)->nof;
	for (fb = f2, n2 = 0; *fb; fb++)
		n2 += (*fb)->nof;

	// determine crossover points
/*
	c1 = random() % n1;
	c2 = random() % n2;
*/
	c1 = c2 = random() % MIN(n1,n2);


	//	fprintf(stderr, "sym count: %d %d, crossing at %d %d\n", n1, n2, c1, c2);

	// copy new strands up to crossed fraglet
	fa = calloc(1, sizeof(struct frag **));
	ca = 0;
	for (fx = f1; *fx && c1 >= (*fx)->nof; fx++) {
		c1 -= (*fx)->nof;
		fa[ca++] = *fx;
		fa = realloc(fa, (ca+1) * sizeof(struct frag*));
		fa[ca] = NULL;
	}
	fb = calloc(1, sizeof(struct frag **));
	cb = 0;
	for (fy = f2; *fy && c2 >= (*fy)->nof; fy++) {
		c2 -= (*fy)->nof;
		fb[cb++] = *fy;
		fb = realloc(fb, (cb+1) * sizeof(struct frag*));
		fb[cb] = NULL;
	}

	// handle cross points in the middle of fraglets
	// (if one cross point is at a boundary, we force the other too)
	if (c1 && c2) {
		f = calloc(1, sizeof(struct frag));
		f->nof = c1 + (*fy)->nof - c2;
		f->sym = malloc(f->nof * sizeof(char*));
		memcpy(f->sym, (*fx)->sym, c1 * sizeof(char*));
		memcpy(f->sym + c1, (*fy)->sym + c2,
		       ((*fy)->nof - c2) * sizeof(char*));
		fa[ca++] = f;
		fa = realloc(fa, (ca+1) * sizeof(struct frag*));
		fa[ca] = NULL;

		f = calloc(1, sizeof(struct frag));
		f->nof = c2 + (*fx)->nof - c1;
		f->sym = malloc(f->nof * sizeof(char*));
		memcpy(f->sym, (*fy)->sym, c2 * sizeof(char*));
		memcpy(f->sym + c2, (*fx)->sym + c1,
		       ((*fx)->nof - c1) * sizeof(char*));
		fb[cb++] = f;
		fb = realloc(fb, (cb+1) * sizeof(struct frag*));
		fb[cb] = NULL;

		fx++;
		fy++;
	}

	// copy cross strands after crossed fraglet
	while (*fy) {
		fa[ca++] = *fy;
		fa = realloc(fa, (ca+1) * sizeof(struct frag*));
		fa[ca] = NULL;
		fy++;
	}
	while (*fx) {
		fb[cb++] = *fx;
		fb = realloc(fb, (cb+1) * sizeof(struct frag*));
		fb[cb] = NULL;
		fx++;
	}

/*
	for (fx = fa; *fx; fx++) {
		f = *fx;
		printf("a)  %s [", f->sym[0]);
		for (i = 1; i < f->nof; i++)
			printf( i == 1 ? "%s" : ":%s", f->sym[i]);
		printf("]\n");
	}
	for (fx = fb; *fx; fx++) {
		f = *fx;
		printf("b)  %s [", f->sym[0]);
		for (i = 1; i < f->nof; i++)
			printf( i == 1 ? "%s" : ":%s", f->sym[i]);
		printf("]\n");
	}
*/
	output(fa);
	output(fb);
}

int main(int argc, char *argv[])
{
	char buf[10000];
	int debug = 0;
	int lim = 1;

	if (argv[1] && !strcmp(argv[1], "-d")) {
		debug = 1;
		argv++,	argc--;
	}

	if (argv[1] && atoi(argv[1]) > 0) {
		lim = atoi(argv[1]);
		argv++,	argc--;
	}

	{
		struct timeb t;

		ftime(&t);
		srandom(t.time ^ t.millitm );
	}

	while (fgets(buf, sizeof(buf)-1, stdin)) {
		if (tolower(buf[0]) != 'g')
			continue;
		parse_genotype(buf);
	}

	while (lim-- > 0) {
		int i, j;
		i = random() % gtcnt;
		j = (i + 1 + random() % (gtcnt - 1)) % gtcnt;
//		fprintf(stderr, "crossing %d and %d\n", i, j);
		docross(genotypes[i], genotypes[j]);
	}

	return 0;
}
