/* fraglets0.15/src/rgen.c

  Copyright (C) 2003-2004 <christian.tschudin@unibas.ch>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING
*/


/*

Function: create a random system description.

Parameters:   

Input (stdin):

  line 1:  list of node names e.g.
    a b c

  line 2:  four parameters - # of genotypes, max fraglets per node,
                          max fraglet length, max multiplicity, e.g.
    100 3 5 2

  line 2 and higher: weight and symbol sequence, e.g.
    10 match
    10 hello
    10 send:seg:a
    10 _
    60 $20

  (the $20 format says: put a tag with number in the range 0..19, e.g. t11)
  (In this case: in 6 of 10 cases we insert a tag)

  or fraglets always to be included:

   f a[...]4

Output: (stdout):

  one genotype per line, internally sorted by nodes and fraglets

  g /a[...]2/.../

*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/timeb.h>

struct mut {
	char *cp;
	int w;
	struct mut *next;
} *mutlist;

struct frag {
	int len;
	int mult;
	struct node *nd;
	struct frag *next;
	char *cp;
	char **sym;
} *fraglist, *inclist;

struct node {
	char *cp;
	struct node *next;
} *nodelist;


static int fragcmp(const void *f1, const void *f2)
{
	return strcmp((*(struct frag **)f1)->cp, (*(struct frag **)f2)->cp);
}


int main(int argc, char *argv[])
{
	char ns[1000], ms[512], *cp;
	int maxg, maxf, maxl, maxm;
	int tot, r;
	struct mut *m;

	{
		struct timeb t;

		ftime(&t);
		srandom(t.time ^ t.millitm );
	}

	// line 1, loop through node names
	fgets(ns, sizeof(ns)-1, stdin);
	for (cp = strtok(ns, " \t\n"); cp; cp = strtok(NULL, " \t\n")) {
		struct node *nd = malloc(sizeof(struct node));
		nd->cp = cp;
		nd->next = nodelist;
		nodelist = nd;
	}

	// line 2, get params
	if (4 != scanf("%d %d %d %d\n", &maxg, &maxf, &maxl, &maxm)) {
		fprintf(stderr, "error on line 2\n");
		exit(-1);
	}

	// line 3+, parse mutation probabilties
	tot = 0;
	while (fgets(ms, sizeof(ms)-1, stdin)) {
		if (ms[0] == 'f') {
			struct frag *f = calloc(1, sizeof(*f));
			f->cp = malloc(strlen(ms + 1));
			strcpy(f->cp, strtok(ms+1, " \t\n"));
			f->next = inclist;
			inclist = f;
		} else {
			int i;

			if (2 != sscanf(ms, "%d %25s\n", &i, &ms))
				continue;
			m = malloc(sizeof(*m));
			m->w = i;
			tot += i;
			m->cp = malloc(strlen(ms) + 1);
			strcpy(m->cp, ms);
			m->next = mutlist;
			mutlist = m;
		}
	}

	// generate the genotypes
	while (maxg-- > 0) {
		struct node *n;
		int i, nof;

		nof = 0;

		// add the "always included" fraglets
		if (inclist) {
			struct frag *f, *fi;

			for (fi = inclist; fi; fi = fi->next) {
				f = calloc(1, sizeof(*f));
				f->cp = malloc(strlen(fi->cp)+1);
				strcpy(f->cp, fi->cp);
				f->next = fraglist;
				fraglist = f;
				nof++;
			}
		}

		// loop through the nodes
		for (n = nodelist; n; n = n->next) {
			int j, len;

			// fprintf(stderr, "rgen for node %s\n", n->cp);

			// loop through fraglets
			i = 1 + random() % (maxf-1);
			nof += i;
			for (; i > 0; i--) {
				struct frag *f;

				// loop through symbols
				len = 1 + random() % maxl;
				f = calloc(1, sizeof(*f));
				f->len = len;
				f->mult = 1 + random() % maxm;
				f->sym = malloc(len * sizeof(char *));
				f->nd = n;
				f->next = fraglist;
				fraglist = f;
				
				for (j = 0; j < len; j++) {
					// choose replacement symbol
					r = random() % tot;
					for (m = mutlist; r > m->w; m = m->next) {
						r -= m->w;
					}
					f->sym[j] = m->cp;
				}
			}
		}


		{
			struct frag *f, **fa;

			fa = malloc(nof * sizeof(struct frag*));

			for (f = fraglist, i=0; f; f = f->next, i++) {
				fa[i] = f;
				if (!f->cp) {
					int j;
					char s[10000];
					int l;
					l = 0;
					l += sprintf(s+l, "%s[", f->nd->cp);
					for (j = 0; j < f->len; j++) {
						char *fs = j ? ":" : "";
						if (f->sym[j][0] == '$')
							l += sprintf(s+l, "%st%d", fs,
								     (int)(random() % atoi(f->sym[j] + 1)));
						else
							l += sprintf(s+l, "%s%s", fs, f->sym[j]);
					}
					l += sprintf(s+l, "]%d", f->mult);
					f->cp = malloc(l + 1);
					strcpy(f->cp, s);
				}
			}

//			qsort(fa, nof, sizeof(struct frag *), fragcmp);

			printf("g ");
			for (i = 0; i < nof; i++) {
				printf("/%s", fa[i]->cp);
				free(fa[i]->cp);
				if (fa[i]->sym)
					free(fa[i]->sym);
				free(fa[i]);
			}
			printf("/\n");

			free(fa);
			fraglist = 0;
		}
	}
	return 0;
}
