/* fraglets0.18/src/sys-send-deliver.c

  sample application that interfaces with the fraglet interpreter

  Copyright (C) 2003-2005 <christian.tschudin@unibas.ch>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING

  This application is composed of a source and a sink side.
  The source side injects payload chunks into the fraglets interpreter
  in the form of [data <payloadsymbol>] fraglets.
  The sink side consumes this chunks and acknowledges their reception.

  The delivery of fraglets from the interpreter to the application
  occurs via the "_" trigger.

  The application expects a reliable delivery from the underlying
  protocol (executed by the fraglet interpreter). It evaluates the
  protocol according to a fitness function that measures how well the
  protocol performed as expected.

  Fitness evaluation is performed in a centralized way, i.e. source
  and sink communicate via off-line means to evaluate the performance
  of the underlying protocol implementation running in the fraglets
  interpreter. [see selfman'05 or wac'05]

  History:
  20050221  lidia.yamamoto@unibas.ch: readjusted fitness function to
            penalize further for resource consumption; fixed trace
            printout (fraglets v0.18)
  20050129  lidia.yamamoto@unibas.ch: reorganized code to have clear
            fitness evaluation points at source and sink; redefined
            fitness criteria (fraglets v0.17)
  20050128  christian.tschudin@unibas.ch: readjusted traces and score
            for selfman'05 (fraglets v0.16)
  20040704  updated by christian.tschudin@unibas.ch for wac'04 (fraglets v0.15)
  200307??  created by christian.tschudin@unibas.ch (ains'03)

*/

#include "frag.h"
#include "frag.c"


// -----------------------------------------------------------------

static int score = 0;
static int sn = 1;
static int nextSN = 1;

static void newPayload(symbol nd, int seq)
{
	struct fragbuf *b;
	char payload[100];

	b = calloc(1, sizeof(*b) + sizeof(symbol) * 2);
	b->len = 2;
	b->data[0] = name_add("data");
	sprintf(payload, "p%d", seq);
	b->data[1] = name_add(payload);

	rx_fraglet(f_new(b, 0, 2, 1, nd));
}


int appl_clockEvent(struct node *n, int clock)
{
#ifdef XXX
	if ( (clock % 25) != 1 ||
	     strcmp(name_lookup(n->name), "a") )
		return 0;

	DEBUG(8, fprintf(stderr, "# clock event %d\n", clock););

	node_lookup(n->name, 0)->symbalance += 2;
	newPayload(n->name, sn++);
#endif
	return 1;
}


// fitness evaluation at node A (source node)
//
void node_a_fitness(struct frag *f) {
  symbol s;

  if (f->len == 2 &&
      f->buf->data[f->offs + 1] == name_add("ack") &&
      sn+1 == nextSN ) {
    // good ack received, reward it with high score
    score += 200;
    sn++;
    DEBUG(2, fprintf(stderr, "node a: good ack, adding [data:p%d] (score %d)\n", sn, score););
    newPayload(f->nd, sn);
    node_lookup(f->nd, 0)->symbalance -= 2;
  } else {
    // bad ack received, penalty is proportionally to amount of trash
    DEBUG(2, fprintf(stderr, "node a: bad ack flen=%d (score %d)\n", f->len, score););
    score -= 50 * ( f->len > 0 ? f->len : 1 );
  }
}


// fitness evaluation at node B (sink node)
//
void node_b_fitness(struct frag *f) {
  symbol s;
  if (f->len > 1) {
    char *cp, expected[100];
    s = f->buf->data[f->offs + 1];
    cp = name_lookup(s);

    DEBUG(2, fprintf(stderr, "node b: payload rcvd, flen=%d", f->len););
    sprintf(expected, "p%d", nextSN);

    // Lidia: expected length must match, as well
    // as the expected symbol
    if (f->len == 2 && !strcmp(expected, cp)) {
      struct fragbuf *b;
      score += 200;
      DEBUG(2, fprintf(stderr, ", correct %s (score %d), adding [ack]\n", cp, score););
      nextSN++;
      node_lookup(f->nd, 0)->symbalance -= 1;
      b = calloc(1, sizeof(*b) + sizeof(symbol));
      b->len = 1;
      b->data[0] = name_add("ack");
      rx_fraglet(f_new(b, 0, 1, 1, f->nd));
    } else {
      // Lidia: decrement score by the amount of
      // trash received
      score -= 50 * f->len ;
      DEBUG(2, fprintf(stderr, ", wrong (score %d)\n", score););
    }
  } else {
    score -= 50;
    DEBUG(2, fprintf(stderr, "node b: trash rcvd, flen=%d (score %d)\n", f->len, score););
  }
}


void appl_trigger(struct frag *f) {
  node_lookup(f->nd, 0)->symbalance += f->len;

  DEBUG(2, fprintf(stderr, "# %05d  local trigger: ", node_lookup(f->nd, 0)->opcnt););

  if (f->nd == name_add("a")) {
    node_a_fitness(f);
  } else if (f->nd == name_add("b")) {
    node_b_fitness(f);
  } else {
    DEBUG(2, fprintf(stderr, "wrong node\n"););
  }
  f_free(f);
}


int flist_memory(struct frag *f)
{
	int m = 0;

	while (f) {
		m += f->mult * f->len;
		f = f->next;
	}

	return m;
}


int node_memory(struct node *n)
{
	int m = 0, i;

	m += flist_memory(n->tq.first);
	for (i = 0; i < MAXSYMBOLS; i++) {
		m += flist_memory(n->mtab[i].first);
		m += flist_memory(n->otab[i].first);
	}
	return m;
}

int appl_finalScore()
{
	struct node *n;

	for (n = nodelist; n; n = n->next) {
//		score -= 3 * n->opcnt;
//		score += n->symbalance;
	}
	DEBUG(2, fprintf(stderr, "final score: %d\n", score););
	return score;
}


void appl_scoreReset()
{
	struct node *n;
	int m;

	score = 0;
	sn = 0;
	nextSN = 1;

	// take initial genotype size into account

	for (n = nodelist; n; n = n->next) {
		n->symbalance = 0;
		m = node_memory(n);
//		fprintf(stderr, "memory %d\n", m);
		// Lidia: penalise further for the size:
		//		score -=  m / 2;
		score -=  m * 100;
	}
	DEBUG(2, fprintf(stderr, "initial score (length): %d\n", score););
}
